import logging
from collections import defaultdict
from dataclasses import dataclass, asdict

log = logging.getLogger(__name__)


class Observable:
    # A set of all attributes which get changed
    changed_attributes = set()

    def __init__(self):
        self.observed = defaultdict(list)

    def __setattr__(self, name, value):
        super().__setattr__(name, value)

        for observer in self.observed.get(name, []):
            observer(name)

    def add_observer(self, name):
        self.observed[name].append(lambda name: self.changed_attributes.add(name))


def sanitise_args(obj: object, args, kwargs):
    kwarg_dict = dict()
    obj.changed_attributes = set()

    for arg in args:
        if isinstance(arg, dict):
            kwarg_dict.update(arg)

    for k, v in kwargs.items():
        if hasattr(type(obj), k):
            kwarg_dict[k] = v
            obj.changed_attributes.add(k)
        else:
            log.debug(f"{obj.__class__.__name__}.{k} - attribute ignored.")
            # raise ValueError(f"No such attribute: {k}")

    return kwarg_dict


@dataclass
class HikDevice:
    ISAPIParams: any = None
    activeStatus: bool = False
    devIndex: str = ""
    devMode: str = ""
    devName: str = ""
    devStatus: str = ""
    devType: str = ""
    devVersion: str = ""
    protocolType: str = ""
    videoChannelNum: int = -1,

    def __init__(self, *args, **kwargs):
        super().__init__()
        device_dict = sanitise_args(self, args, kwargs)

        for property_name in device_dict:
            if isinstance(device_dict[property_name], (str, type(None), bool, int)):
                setattr(self, property_name, device_dict[property_name])

            if isinstance(device_dict[property_name], dict):
                if property_name == "ISAPIParams":
                    self.ISAPIParams = device_dict[property_name]

    def dict(self):
        controller_dict = {}
        for k, v in asdict(self).items():
            if isinstance(v, (list, dict, bool, int)):
                controller_dict[k] = v
            elif isinstance(v, type(None)):
                controller_dict[k] = None
            else:
                controller_dict[k] = str(v)

        return controller_dict



import json
import os
from datetime import datetime
import filetype
from .hik_dataclasses import HikDevice
from .hik_error import HikDeviceGatewayError


class HikAccessControl:

    def add_person(self, device: HikDevice, employee_num: str, name: str, begin_time: datetime, end_time: datetime):
        if begin_time.replace(second=0, microsecond=0).isoformat() == end_time.replace(second=0,
                                                                                       microsecond=0).isoformat():
            raise HikDeviceGatewayError('Begin and End DateTimes cannot be the same.')

        raw = '''{
            "UserInfo": [
                {
                    "employeeNo": "",
                    "name": "",
                    "Valid": {
                        "beginTime": "2017-01-01T00:00:00",
                        "endTime": "2027-12-31T23:59:59"
                    }
                }
            ]
        }'''
        json_body = json.loads(raw)
        json_body['UserInfo'][0]['employeeNo'] = employee_num
        json_body['UserInfo'][0]['name'] = name
        json_body['UserInfo'][0]['Valid']['beginTime'] = begin_time.replace(second=0, microsecond=0).isoformat()
        json_body['UserInfo'][0]['Valid']['endTime'] = end_time.replace(second=0, microsecond=0).isoformat()

        url_path_params = f"/ISAPI/AccessControl/UserInfo/Record?format=json&devIndex={device.devIndex}"
        response = self.json_query(
            "POST",
            self.baseurl + url_path_params,
            json_body)

        if 'errorMsg' in response:
            raise HikDeviceGatewayError(response['errorMsg'])

        if 'UserInfoOutList' not in response:
            raise HikDeviceGatewayError("'UserInfoOutList' missing from response body.")

        if 'UserInfoOut' not in response['UserInfoOutList']:
            raise HikDeviceGatewayError("'UserInfoOut' missing from response body.")

        if not isinstance(response['UserInfoOutList']['UserInfoOut'], list):
            raise HikDeviceGatewayError("'UserInfoOut' is not a list.")

        if len(response['UserInfoOutList']['UserInfoOut']) == 0:
            return []

        if 'statusCode' not in response['UserInfoOutList']['UserInfoOut'][0]:
            raise HikDeviceGatewayError("'statusCode' missing from response body.")

        if response['UserInfoOutList']['UserInfoOut'][0]['statusCode'] != 1:
            if 'errorMsg' in response['UserInfoOutList']['UserInfoOut'][0]:
                raise HikDeviceGatewayError(response['UserInfoOutList']['UserInfoOut'][0]['errorMsg'])
            else:
                raise HikDeviceGatewayError(
                    f'Error - StatusCode: {str(response['UserInfoOutList']['UserInfoOut'][0]['statusCode'])}')

        return True

    def delete_person(self, device: HikDevice, employee_num: str):
        raw = '''{
            "UserInfoDetail": {
                "mode": "byEmployeeNo",
                "EmployeeNoList": [
                    {
                        "employeeNo": "123456"
                    }
                ]
            }
        }'''
        json_body = json.loads(raw)
        json_body['UserInfoDetail']['EmployeeNoList'][0]['employeeNo'] = employee_num

        url_path_params = f"/ISAPI/AccessControl/UserInfoDetail/Delete?format=json&devIndex={device.devIndex}"
        response = self.json_query(
            "PUT",
            self.baseurl + url_path_params,
            json_body)

        if 'statusCode' not in response:
            raise HikDeviceGatewayError("'statusCode' missing from response body.")

        if response['statusCode'] != 1:
            if 'errorMsg' in response:
                raise HikDeviceGatewayError(response['errorMsg'])
            else:
                raise HikDeviceGatewayError(f'Error - StatusCode: {str(response['statusCode'])}')

        return True

    def add_card(self, device: HikDevice, employee_num: str, card_num: str):
        raw = '''{
            "CardInfo": {
                "employeeNo": "123456",
                "cardNo": "1234567890"
            }
        }'''
        json_body = json.loads(raw)
        json_body['CardInfo']['employeeNo'] = employee_num
        json_body['CardInfo']['cardNo'] = card_num

        url_path_params = f"/ISAPI/AccessControl/CardInfo/Record?format=json&devIndex={device.devIndex}"
        response = self.json_query(
            "POST",
            self.baseurl + url_path_params,
            json_body)

        if 'statusCode' not in response:
            raise HikDeviceGatewayError("'statusCode' missing from response body.")

        if response['statusCode'] != 1:
            if 'errorMsg' in response:
                raise HikDeviceGatewayError(response['errorMsg'])
            else:
                raise HikDeviceGatewayError(f'Error - StatusCode: {str(response['statusCode'])}')

        return True

    def delete_card(self, device: HikDevice, card_num: str):
        raw = '''{
            "CardInfoDelCond": {
                "CardNoList": [
                    {
                        "cardNo": "1234567890"
                    }
                ]
            }
        }'''
        json_body = json.loads(raw)
        json_body['CardInfoDelCond']['CardNoList'][0]['cardNo'] = card_num

        url_path_params = f"/ISAPI/AccessControl/CardInfo/Delete?format=json&devIndex={device.devIndex}"
        response = self.json_query(
            "PUT",
            self.baseurl + url_path_params,
            json_body)

        if 'statusCode' not in response:
            raise HikDeviceGatewayError("'statusCode' missing from response body.")

        if response['statusCode'] != 1:
            if 'errorMsg' in response:
                raise HikDeviceGatewayError(response['errorMsg'])
            else:
                raise HikDeviceGatewayError(f'Error - StatusCode: {str(response['statusCode'])}')

        return True

    def add_face(self, device: HikDevice, employee_num: str, face_img: str):
        if not os.path.isfile(face_img):
            raise HikDeviceGatewayError(f'Failed to find {face_img}')

        if not filetype.is_image(face_img):
            raise HikDeviceGatewayError(f'{face_img} is not an image')

        raw = '''{
            "FaceInfo": {
                "employeeNo": "123456"
            }
        }'''

        json_body = json.loads(raw)
        json_body['FaceInfo']['employeeNo'] = employee_num

        files = {
            'data': (None, json.dumps(json_body).encode('ascii'), 'application/json'),
            'FaceDataRecord':(None, open(face_img, 'rb'), 'image/jpeg')
        }

        url_path_params = f"/ISAPI/Intelligent/FDLib/FaceDataRecord?format=json&devIndex={device.devIndex}"
        response = self.json_query(
            "POST",
            self.baseurl + url_path_params,
            None,
            files=files)

        if 'statusCode' not in response:
            raise HikDeviceGatewayError("'statusCode' missing from response body.")

        if response['statusCode'] != 1:
            if 'errorMsg' in response:
                if 'subStatusCode' in response:
                    raise HikDeviceGatewayError(f"{response['errorMsg']} - {response['subStatusCode']}")
                raise HikDeviceGatewayError(response['errorMsg'])
            else:
                raise HikDeviceGatewayError(f'Error - StatusCode: {str(response['statusCode'])}')

        return True

    def delete_face(self, device: HikDevice, employee_num: str):
        raw = '''{
            "FaceInfoDelCond": {
                "EmployeeNoList": [
                    {
                        "employeeNo": "123456"
                    }
                ]
            }
        }'''
        json_body = json.loads(raw)
        json_body['FaceInfoDelCond']['EmployeeNoList'][0]['employeeNo'] = employee_num

        url_path_params = f"/ISAPI/Intelligent/FDLib/FDSearch/Delete?format=json&devIndex={device.devIndex}"
        response = self.json_query(
            "PUT",
            self.baseurl + url_path_params,
            json_body)

        if 'statusCode' not in response:
            raise HikDeviceGatewayError("'statusCode' missing from response body.")

        if response['statusCode'] != 1:
            if 'errorMsg' in response:
                raise HikDeviceGatewayError(response['errorMsg'])
            else:
                raise HikDeviceGatewayError(f'Error - StatusCode: {str(response['statusCode'])}')

        return True



import binascii
import time
import base64
from urllib.parse import urlparse


def url_parser(url):
    parts = urlparse(url)
    directories = parts.path.strip('/').split('/')
    queries = parts.query.strip('&').split('&')
    host = parts.netloc.strip(':').split(':')[0]

    elements = {
        'scheme': parts.scheme,
        'host': host,
        'path': parts.path,
        'params': parts.params,
        'query': parts.query,
        'port': parts.port,
        'fragment': parts.fragment,
        'directories': directories,
        'queries': queries,
    }

    return elements


class ConvertBase64:

    @staticmethod
    def encode(text: str):
        return base64.b64encode(text.encode('ascii')).decode('ascii')

    @staticmethod
    def decode(text: str):
        try:
            return base64.b64decode(text.encode('ascii')).decode('ascii')
        except binascii.Error:
            return base64.b64decode(text.encode('ascii') + b"==").decode('ascii')






class Stopwatch:

    def __init__(self):
        self._time = 0

    def start(self):
        self._time = time.time()
        return self

    def stop(self):
        self._time = time.time() - self._time

    def print(self, unit=None, precision=2, show_unit=1):
        division_table = {'h': 360, 'm': 60, 's': 1, 'ms': 0.001}
        unit_table = {'ms': "milliseconds", 's': "seconds", 'm': "minutes", 'h': "hours"}
        if unit not in division_table:
            for key, val in division_table.items():
                if self._time >= val:
                    unit = key
                    break
            else:
                unit = 'ms'
        unit_text = ""
        if show_unit == 1:
            unit_text = " " + unit
        elif show_unit == 2:
            unit_text = " " + unit_table[unit]
        elif show_unit == 3:
            unit_text = " " + unit_table[unit].capitalize()
        return f"{(self._time / division_table[unit]):.{precision}f}{unit_text}"

    def __str__(self):
        return str(self._time)



import json

from .hik_dataclasses import HikDevice
from .hik_error import HikDeviceGatewayError


class HikDGDeviceManagement():

    def list_devices(self):
        raw = '''{
            "SearchDescription": {
                "position": 0,
                "maxResult": 100
            }
        }'''
        json_body = json.loads(raw)
        response = self.json_query("POST", self.baseurl + "/ISAPI/ContentMgmt/DeviceMgmt/deviceList?format=json", json_body)

        if 'SearchResult' not in response:
            raise HikDeviceGatewayError("'SearchResult' missing from response body.")

        response = response['SearchResult']

        if 'numOfMatches' not in response:
            raise HikDeviceGatewayError("'numOfMatches' missing from response body.")

        if response['numOfMatches'] == 0:
            return []

        if 'MatchList' not in response:
            raise HikDeviceGatewayError("'MatchList' missing from response body.")

        if not isinstance(response['MatchList'], list):
            raise HikDeviceGatewayError("'MatchList' is not a list.")

        device_list = []
        for entry in response['MatchList']:
            if 'Device' in entry:
                device_list.append(HikDevice(entry['Device']))

        return device_list



import logging
import json
import os
import ssl
from enum import Enum
from json import JSONDecodeError
from pathlib import Path
import requests
from requests.auth import HTTPDigestAuth

log = logging.getLogger(__name__)


class HikDGConnection:

    def __init__(self, url_components, user, pwd):
        self.ssl_context = None
        self.url_components = url_components
        self.user = user
        self.pwd = pwd
        if url_components['host'] == '':
            raise ValueError("Invalid Connection URL")
        if url_components['scheme'] == '':
            url_components['scheme'] = 'http'

        self.baseurl = f"{url_components['scheme']}://{url_components['host']}"
        if url_components['port']:
            self.baseurl = self.baseurl + f":{url_components['port']}"

        log.info(f"HikDevice Gateway BASE_URL: {self.baseurl}")

    def json_query(self, method, url, json_body: dict = '', headers=None, files=None):
        raw_body = ''
        if json_body:
            if not isinstance(json_body, dict):
                raise ValueError("Variable 'json_body' must be of type dict.")
            else:
                raw_body = json.dumps(json_body)

        headers = headers or {
            'Content-Type': 'application/json',
            'Accept': 'application/json'
        }

        log.debug(f"Request data: host={self.baseurl}, {method}, {url}, {headers}, {raw_body}")

        if isinstance(files, dict):
            headers = None
            raw_body = None

        response = requests.request(method,
                                    url,
                                    data=raw_body,
                                    headers=headers,
                                    auth=HTTPDigestAuth(self.user, self.pwd),
                                    files=files)

        # Try to convert body into json
        try:
            json_body = json.loads(response.text)
        except JSONDecodeError:
            json_body = None
        except Exception as e:
            log.error(e)
            json_body = None

        return json_body


def pretty_print_POST(req):
    """
    At this point it is completely built and ready
    to be fired; it is "prepared".

    However pay attention at the formatting used in
    this function because it is programmed to be pretty
    printed and may differ from the actual request.
    """
    print('{}\n{}\r\n{}\r\n\r\n{}'.format(
        '-----------START-----------',
        req.method + ' ' + req.url,
        '\r\n'.join('{}: {}'.format(k, v) for k, v in req.headers.items()),
        req.body,
    ))



from setuptools import setup, find_packages

long_description = open('README.rst').read()

setup(name="pyHikUp",
      packages=find_packages(),
      version="0.0.1",
      author="John Owen",
      description="Python wrapper for Hik Device Gateway",
      long_description_content_type='text/markdown',
      long_description=long_description,
      maintainer_email="sales@sensoraccess.co.uk",
      install_requires=['validators', 'fuzzywuzzy', 'cryptography', 'pysignalr', 'websockets', 'python-Levenshtein'],
      #packages=['pyGuardPoint'],
      license_files=('LICENSE.txt',),
      zip_safe=False)
